class SWEnv:
    task_id = "SW_TASK_ID"
    job_id = "SW_JOB_ID"
    status_dir = "SW_TASK_STATUS_DIR"
    log_dir = "SW_TASK_LOG_DIR"
    result_dir = "SW_TASK_RESULT_DIR"
    input_config = "SW_TASK_INPUT_CONFIG"
